/**
 * base64 streamline inplace decoder.
 *
 * Features / assumptions:
 * - optimized uint32 read/write (only LE support!)
 * - lazy chunkwise decoding
 * - errors out on any non base64 chars (no support for NL formatted base64)
 * - decodes in wasm
 * - inplace decoding to save memory
 * - supports a keepSize for lazy memory release
 */
export default class Base64Decoder {
    keepSize: number;
    private _d;
    private _m32;
    private _inst;
    private _mem;
    constructor(keepSize: number);
    /**
     * Currently decoded bytes (borrowed).
     * Must be accessed before calling `release` or `init`.
     */
    get data8(): Uint8Array;
    /**
     * Release memory conditionally based on `keepSize`.
     * If memory gets released, also the wasm instance will be freed and recreated on next `init`,
     * otherwise the instance will be reused.
     */
    release(): void;
    /**
     * Initializes the decoder for new base64 data.
     * Must be called before doing any decoding attempts.
     * `size` is the max amount of decoded bytes to be expected.
     * The method will either spawn a new wasm instance or grow
     * the needed memory of an existing instance.
     */
    init(size: number): void;
    /**
     * Put bytes in `data` into the decoder.
     * Additionally decodes the payload, if it reached 2^17 bytes.
     * Returns 1 on error, else 0.
     */
    put(data: Uint8Array | Uint16Array | Uint32Array): number;
    /**
     * End the current decoding.
     * Also decodes leftover payload from previous put calls.
     * Returns 1 on error, else 0.
     */
    end(): number;
}
